unit MyUtils;
  (* Little routines of general utility *)

  interface
    uses
      Buttons    (* tBitBtn *),
      Classes    (* tComponent, tList, tNotifyEvent, tShiftState *),
      ComCtrls   (* tCustomRichEdit, tProgressBar *),
      DB         (* tDataSet, dsEdit, dsInsert, ftInteger, tIntegerField *),
      DBGrids    (* tCustomDBGrid *),
      DBTables   (* tDatabase, tQuery, tTable, ttDefault *),
      Forms      (* tDataModule, tForm, tFormClass, Screen *),
      QuickRpt   (* tQuickRep *),
      Windows    (* CreateMutex, DWORD, ERROR_ALREADY_EXISTS, GetLastError,
                    HKEY_LOCAL_MACHINE, Keybd_Event, KEYEVENTF_EXTENDEDKEY,
                    KEYEVENTF_KEYUP, MapVirtualKey, SendMessage, SW_SHOWNORMAL,
                    VK_CONTROL, VK_MENU, VK_SHIFT *);

    const
      DefaultShrinkAway = false;

    type
      dtformat = (dtNONE, dtSAMPLE,
                    dtMDYY,   dtMDYYYY,
                    dtMMDDYY, dtMMDDYYYY,
                    dtYYYYMD, dtYYYYMMDD);
      tTableAction = (TABLE_CLOSE, TABLE_OPEN, TABLE_REFRESH);

    type
      tResponsibleList = class(tList)
        (* clean up components *)
        public
          procedure Clear;    override;
          destructor Destroy; override;
          function Dump: string; virtual;
        end (* tResponsibleList *);

      tResponsibleObject = class(tObject)
        protected
          function Dump(const N: integer): string; virtual; abstract;
        end (* tResponsibleObject *);

      tRRFCloseBtn = class(tBitBtn)
        private
          fShrinkAway: boolean;
          fShrinkRate: integer;
          fOnBeforeClose: tNotifyEvent;
          procedure CloseForm(Sender: tObject);
          procedure SetOnBeforeClose(const Value: tNotifyEvent);
          procedure SetShrinkRate(const Value: integer);
        protected
          procedure Loaded; override;
        public
          constructor Create(AnOwner: tComponent); override;
        published
          property ShrinkAway:    boolean      read    fShrinkAway
                                               write   fShrinkAway
                                               default DefaultShrinkAway;
          property ShrinkRate:    integer      read    fShrinkRate
                                               write   SetShrinkRate
                                               default 3;
          property OnBeforeClose: tNotifyEvent read    fOnBeforeClose
                                               write   SetOnBeforeClose;
        end (* tRRFCloseBtn *);
        (* eject *)
      tRRFDTPicker = class(tDateTimePicker)
        private
          fRRFDTFormat: dtformat;
          fRRFDTAnyFormat: string;
          procedure SetRRFDTFormat(const Value: dtformat);
          procedure SetRRFDTAnyFormat(const Value: string);
        public
          constructor Create(AnOwner: tComponent); override;
        published
          property RRFDTAnyFormat: string   read    fRRFDTAnyFormat
                                            write   SetRRFDTAnyFormat;
          property RRFDTFormat:    dtformat read    fRRFDTFormat
                                            write   SetRRFDTFormat
                                            default dtNONE;
        end (* tRRFDTPicker *);

      tRRFLock = class(tComponent)
        private
          fDataBase    (* locked or holds locked table *): tDatabase;
          LockTable    (* compete to edit this         *): tTable;
          procedure SetDataBase(const Value: tDatabase);
        public
          procedure   Abandon;
          constructor Create(AnOwner: tComponent); override;
          destructor  Destroy;                     override;
          function    Lock: boolean;
          procedure   Unlock;
        published
          property DataBase: tDatabase  read  fDataBase
                                        write SetDataBase;
        end (* tRRFLock *);
        (* eject *)
      tRRFMemo = class(tCustomRichEdit)
        private
          function GetSelectedLine: integer;
          procedure SetSelectedLine(const Value: integer);
        public
          function PosLineStart: integer;
          function SelectedLineIsBlank: boolean;
          property SelectedLine: integer read  GetSelectedLine
                                         write SetSelectedLine;
        published
          property Anchors;
          property Enabled;
          property Font;
          property HideScrollBars;
          property HideSelection;
          property Lines;
          property PlainText;
          property ScrollBars;
          property SelLength;
          property SelStart;
          property SelText;
          property TabOrder;
          property Visible;
          property WantReturns;
          property WantTabs;
          property WordWrap;
          property OnChange;
          property OnClick;
          property OnDblClick;
          property OnEnter;
          property OnExit;
          property OnKeyDown;
          property OnKeyPress;
          property OnKeyUp;
          property OnMouseDown;
        end (* tRRFMemo *);

    type
      tDeleteManager = class(tResponsibleList)
        public
          procedure DeleteRecordAfterCascade(MasterTable: tTable);
          procedure LogDependency(MasterTable, ChildTable: tTable;
                                  const MasterFieldName, ChildFieldName: string);
        end (* tDeleteManager *);

    var
      CountingPages: boolean;
      PageNo, NPages: integer;
        (* eject *)
    function  AskYesNo(const FormCaption, Prompt: string): boolean;
    function  DaysInMonth(const TheYear, TheMonth: integer): integer;
    procedure DoQueriesAndTables(TheDM: tDataModule; const WhatToDo: tTableAction);
    function  GetSystemUserName: string;
    procedure InsistOnLocalShare;
    procedure ModalShow(AFormClass: tFormClass);
    function  MyIntVal(const S: string): integer;
    procedure MyPreview(QR: tQuickRep; PB: tProgressBar);
    function  MyStrToCurr(S: string): currency;
    function  MyStrToFloat(S: string): real;
    function  NewUID(UIDField: tIntegerField): integer;
    procedure PurgeIntField(IntField: tIntegerField; const PurgeValue: integer);
    function  RunningInIDE: boolean;
    procedure RunOnlyOneInstance(const UID: PChar);
    procedure SayLockedOut(DataSet: tDataSet);
    procedure SendKey(Key: Word; const Shift: tShiftState; SpecialKey: boolean);
    procedure SetDefaultValues(Obj: tObject);
    function  TrimPathstr(const S: string; const N: integer): string;
    function  WhiteSpace(const S: string): boolean;

  implementation
    uses
      BDE      (* dbiGetSysConfig, dbiSaveChanges, SysConfig *),
      CommCtrl (* DTM_SETFORMAT *),
      Controls (* Canvas, crHourglass, tCursor, mrYes *),
      Dialogs  (* tOpenDialog *),
      Graphics (* fsBold *),
      Messages (* EM_LINEFROMCHAR, EM_LINEINDEX, EM_LINESCROLL, EM_SCROLLCARET *),
      MyConsts (* BACKSLASH, BLANK, CR, HYPHEN, nullstring, PERIOD, TAB *),
      Registry (* tRegistry *),
      SysUtils (* EncodeDate, Format *),
      TypInfo  (* pPropInfo, pPropList, SetOrdProp, tTypeKind *),
      Yesno    (* tYesNoDlg (Yes/No dialog box) *);

      {$R MYUTILS.RES}

    const
      MysteryMBZ = 0;
      NotUsedMBZ = 0;

    type
      tDeleteRecord = class(tObject)
        (* Record the fact that before deleting a record from MTable with
          FieldByName(MFieldName).Value = X, we must first delete all records
          of STable with FieldByName(SFieldName).Value = X *)
        MTable, STable: tTable;
        MFieldName, SFieldName: string;
        constructor Create(MasterTable, SlaveTable: tTable; const MFN, SFN: string);
        end (* tDeleteRecord *);

      tRRFWrapper = class(tObject)
        (* this exists only in order to turn FlushBuffers from a procedure to
         a method *)
          procedure FlushBuffers(DataSet: tDataSet);
        end (* tRRFWrapper *);
        (* eject *)
    var
      RRFWrapper: tRRFWrapper;

    function AskYesNo(const FormCaption, Prompt: string): boolean;
      (* This will sooner or later be moved out to utility unit *)

      begin (* AskYesNo *)
        with tYesNoDlg.Create(nil) do
          try     (* with *)
            Caption := FormCaption;
            PromptLabel.Caption := Prompt;
            Result := (ShowModal = mrYes)
          finally (* with *)
            Free
          end     (* with *)
       end   (* AskYesNo *);

    function CleanForConversion(S: string): string;
      (* service routine for MyStrToCurr and MyStrToFloat *)

      var
        I, PosPeriod: integer;
        Sign: string;

      begin (* CleanForConversion *)
        Sign := nullstring;
        S := Trim(S);
        I := 1;
        PosPeriod := 0;
        if pos(HYPHEN, S) = 1 then
          begin (* minus *)
            Sign := HYPHEN;
            S := copy(S, 2, maxint)
          end   (* minus *);

        while I <= length(S) do
          begin (* look at S[I] *)
            case S[I] of
              '0' .. '9': (* ignore *);

                  PERIOD: if PosPeriod > 0 then
                            break
                          else
                            PosPeriod := I;

                     else break
              end (* case *);
            inc(I)
          end   (* look at S[I] *);

        if I = 1 then
          Result := nullstring
        else if PosPeriod = 0 then
          Result := Sign + copy(S, 1, I-1)
        else if (I = 2) and (PosPeriod = 1) then
          Result := nullstring
        else
          Result := Sign + copy(S, 1, I-1)
      end   (* CleanForConversion *);

    function DaysInMonth(const TheYear, TheMonth: integer): integer;
      (* simple utility *)

      begin (* DaysInMonth *)
        if TheMonth = 12 then
          Result := 31
        else
          Result := trunc(EncodeDate(TheYear, TheMonth+1, 1) -
                          EncodeDate(TheYear,  TheMonth,  1))
      end   (* DaysInMonth *);
      (* eject *)
    procedure DoQueriesAndTables(TheDM: tDataModule; const WhatToDo: tTableAction);
      (* Close or open the tables *)

      var
        I: integer;
        TheComponent: tComponent;

      begin (* DoQueriesAndTables *)
        try (* don't leave cursor messed up *)
          Screen.Cursor := crHourGlass;
          for I := 1 to TheDM.ComponentCount do
            begin (* process Ith component *)
              TheComponent := TheDM.Components[I-1];
              if (TheComponent is tQuery) and (WhatToDo = TABLE_CLOSE) then
                tQuery(TheComponent).Close
              else if TheComponent is TTable then
                with tTable(TheComponent) do
                  case WhatToDo of
                      TABLE_CLOSE: begin (* close *)
                                     if State in [dsEdit, dsInsert] then
                                       Post;
                                     Close
                                   end   (* close *);
				   
                       TABLE_OPEN: begin (* open *)
                                     Close;
                                    { IndexDefs.Update;}
                                     AfterPost := RRFWrapper.FlushBuffers;
                                     Open
                                   end   (* open *);

                    TABLE_REFRESH: Refresh
                    end (* case *)
            end   (* process Ith component *)
        finally (* don't leave cursor messed up *)
          Screen.Cursor := crDefault
        end     (* don't leave cursor messed up *)
      end   (* DoQueriesAndTables *);
      (* eject *)
    function GetSystemUserName: string;
      (* Get user name from the registry *)

      resourcestring
        UserPath = 'Network\Logon';
        UserKey  = 'username';

      const
        CreateIfMissing = true;

      begin (* GetSystemUserName *)
        Result := nullstring;
        with tRegistry.Create do
          try     (* using TheRegistry *)
            RootKey := HKEY_LOCAL_MACHINE;
            if OpenKey(UserPath, not CreateIfMissing) then
              Result := ReadString(UserKey)
          finally (* using TheRegistry *)
            Free
          end     (* using TheRegistry *);
      end   (* GetSystemUserName *);

    procedure InsistOnLocalShare;
      (* from www.itecuk.com/delmag/Paradox.htm *)

      var
        bdeConfig: SysConfig;

      begin (* InsistOnLocalShare *)
        Session.Open;

        if DbiGetSysConfig(bdeConfig) = dbiErr_None then
          if not bdeConfig.bLocalShare then
            begin (* LocalShare not true *)
              ShowMessage('Configuration error.' + CR + CR +
                          '  This application cannot be run until the'    + CR +
                          'BDE Administrator control panel has been used' + CR +
                          'to set the BDE LocalShare flag to True.');
              {$ifdef Win32}
                Application.ShowMainForm := false;
              {$endif}
              Application.Terminate
            end   (* LocalShare not true *)
      end   (* InsistOnLocalShare *);

    procedure ModalShow(AFormClass: tFormClass);
      (* Show a form modally *)

      begin (* ModalShow *)
        with AFormClass.Create(nil) do
          begin (* with *)
            try     (* show it transiently *)
              ShowModal
            finally (* show it transiently *)
              Release
            end     (* show it transiently *)
          end   (* with *)
      end   (* ModalShow *);

    function MyIntVal(const S: string): integer;
      (* Error-free version of IntToStr *)

      var
        I: integer;

      begin (* MyIntVal *)
        Result := 0;
        for I := 1 to length(S) do
          case S[I] of
            '0' .. '9': Result := 10 * Result + ord(S[I]) - ord('0')
          end (* case *)
      end   (* MyIntVal *);

    procedure MyPreview(QR: tQuickRep; PB: tProgressBar);
      (* Preview the quick report, but only after scanning it to get a page
       count *)

      var
        OldCursor: tCursor;

      begin (* MyPreview *)
        with QR do
          begin (* count pages, then preview *)
            OldCursor := Screen.Cursor;
            Screen.Cursor := crHourglass;
            CountingPages := true;
            DataSet.DisableControls;
            NPages := 0;
            try     (* temporize *)
              if PB <> nil then
                with PB do
                  begin (* start progress bar *)
                    Max := DataSet.RecordCount;
                    Position := 0;
                    Visible := true
                  end   (* start progress bar *);
              Prepare;
              NPages := QRPrinter.PageCount;
              QRPrinter.Free;
              QRPrinter := nil;
              PageNo := 0
            finally (* temporize *)
              if PB <> nil then
                PB.Visible := false;
              DataSet.EnableControls;
              CountingPages := false;
              Screen.Cursor := OldCursor
            end     (* temporize *);
            Preview
          end   (* count pages, then preview *)
      end   (* MyPreview *);

    function MyStrToCurr(S: string): currency;
      (* StrToCurr probably fails like StrToFloat *)

      begin (* MyStrToCurr *)
        S := CleanForConversion(S);
        if S = nullstring then
          Result := 0
        else
          Result := StrToCurr(S)
      end   (* MyStrToCurr *);

    function MyStrToFloat(S: string): real;
      (* StrToFloat ought to work, but it doesn't *)

      begin (* MyStrToFloat *)
        S := CleanForConversion(S);
        if S = nullstring then
          Result := 0
        else
          Result := StrToFloat(S)
      end   (* MyStrToFloat *);

    function  NewUID(UIDField: tIntegerField): integer;
      (* Return something bigger than what is there *)

      var
        OldCursor: tCursor;
        TheFieldName, TheTableName: string;

      begin (* NewUID *)
        TheFieldName := UIDField.FieldName;
        TheTableName := (UIDField.DataSet as tTable).TableName;
        OldCursor := Screen.Cursor;
        Screen.Cursor := crHourglass;
        try     (* restore cursor *)
          with tQuery.Create(Application) do
            try     (* reclaim memory *)
              Close;
              DataBaseName := tDBDataSet(UIDField.DataSet).DatabaseName;
              SQL.Clear;
              SQL.Add(Format('Select max(%s) from %s', [TheFieldName, TheTableName]));
              Open;
              Result := Fields[0].AsInteger + 1
            finally (* reclaim memory *)
              Close;
              Free
            end     (* reclaim memory *)
        finally (* restore cursor *)
          Screen.Cursor := OldCursor
        end     (* restore cursor *)
      end   (* NewUID *);

    procedure PurgeIntField(IntField: tIntegerField; const PurgeValue: integer);
      (* Remove records of IntField's table that have the value PurgeValue *)

      begin (* PurgeIntField *)
        with tQuery.Create(nil) do
          try     (* must destroy it *)
            Close;
            DatabaseName := tTable(IntField.DataSet).DatabaseName;
            SQL.Clear;
            SQL.Add(Format('Delete from %s where %s = %d',
                           [(IntField.DataSet as tTable).TableName,
                            IntField.FieldName,
                            PurgeValue]));
            Prepare;
            ExecSQL
          finally (* must destroy it *)
            Free
          end     (* must destroy it *)
      end   (* PurgeIntField *);
      (* eject *)
    function  RunningInIDE: boolean;
      (* Some code doesn't work when program started from within IDE.  Probably
      these problems should all be diagnosed, but in some cases it's easier to
      avoid the situation.  The DebugHook properties are undocumented, but
      reported in one of the Delphi newsgroups *)

      begin (* RunningInIDE *)
        Result := (System.DebugHook <> 0)
      end   (* RunningInIDE *);
      
    procedure RunOnlyOneInstance(const UID: PChar);
      (* Call this in the .DPR file, just before Application.Initialize.
      It tries to create a named mutex that will live as long as the process.
      If the name is in use, the process dies.

        The argument should be an application-unique string, possibly the date &
      time of coding.  A GUID would of course work, but they are ugly.

        The code is taken from inner-smile.com/delphifaq.htm *)

      const
        IgnoreSecurity = nil;
        IgnoreOwnership = false;

      begin (* RunOnlyOneInstance *)
        CreateMutex(IgnoreSecurity, IgnoreOwnership, UID);
        if GetLastError = ERROR_ALREADY_EXISTS then
          Halt
      end   (* RunOnlyOneInstance *);
      (* eject *)
    procedure SayLockedOut(DataSet: tDataSet);
      (* Explain that database locked *)

      begin (* SayLockedOut *)
        ShowMessage('Database locked by another user; try again later')
      end   (* SayLockedOut *);
      
    procedure SendKey(Key: Word; const Shift: tShiftState; SpecialKey: boolean);
      (* from code posted by Peter Below to the borland.public.delphi.winapi
       newsgroup on 21 February 1999.  Sample usage would be
            SendKey(ord('C'), [ssCtrl], false)     copy to clipboard
            SendKey(VK_RETURN, [] false)           new line
       *)

       const
         ExtendedPrefix = KEYEVENTF_EXTENDEDKEY;
         KeyReleased    = KEYEVENTF_KEYUP;
         KeyPressed     = 0;
         VirtualToScan = 0;
         Undocumented = 0;

       type
         byteset = set of 0 .. 7;

         tShiftKeyInfo = record
           Shift: byte;
           VKey: byte;
           end (* tShiftKeyInfo *);

       const
         ShiftKeys: array[1 .. 3] of tShiftKeyInfo =
           ((Shift: ord(ssCtrl);  VKey: VK_CONTROL),
            (Shift: ord(ssShift); VKey: VK_SHIFT  ),
            (Shift: ord(ssAlt);   VKey: VK_MENU));

       var
         BShift: byteset absolute Shift;
         KeyPrefix: DWORD;
         I: integer;
         (* eject *)
       procedure PressKey(const Key: Word; const KeyPrefix: DWORD);
         (* simplify calls to Keybd_Event *)

         begin (* PressKey *)
           Keybd_Event(Key, MapVirtualKey(Key, VirtualToScan), KeyPrefix or KeyPressed,
                         Undocumented)
         end   (* PressKey *);

       procedure ReleaseKey(const Key: Word; const KeyPrefix: DWORD);
         (* simplify calls to Keybd_Event *)

         begin (* ReleaseKey *)
           Keybd_Event(Key, MapVirtualKey(Key, VirtualToScan), KeyPrefix or KeyReleased,
                         Undocumented)
         end   (* ReleaseKey *);
         
       begin (* SendKey *)
         for I := 1 to 3 do
           with ShiftKeys[I] do
             if Shift in BShift then
               PressKey(VKey, 0);

         if SpecialKey then
           KeyPrefix := ExtendedPrefix
         else
           KeyPrefix := 0;

         PressKey  (Key, KeyPrefix);
         ReleaseKey(Key, KeyPrefix);

         for I := 3 downto 1 do
           with ShiftKeys[I] do
             if Shift in BShift then
               ReleaseKey(VKey, 0);
       end   (* SendKey *);
       (* eject *)
    procedure SetDefaultValues(Obj: tObject);
      (* Called from within a class's constructor, this routine sets the
      values of those properties that have default values.  Taken from
      Ray Lischner's Delphi in a Nutshell, page 175. *)

      const
        GetPropCount = nil;
        OrdinalTypes = [tkChar, tkEnumeration, tkInteger, tkSet, tkWChar];
        NoDefault = Low(integer);

      var
        Count, I: integer;
        PropList: pPropList;

      begin (* SetDefaultValues *)
        if Assigned(Obj) then
          if Obj.ClassInfo <> nil then
            begin (* nondegenerate *)
              Count := GetPropList(Obj.ClassInfo, OrdinalTypes, GetPropCount);
              GetMem(PropList, Count * sizeof(pPropInfo));
              try     (* must free this memory *)
                GetPropList(Obj.ClassInfo, OrdinalTypes, PropList);
                for I := 0 to Count - 1 do
                  if PropList[I].Default <> NoDefault then
                    SetOrdProp(Obj, PropList[I], PropList[I].Default)
              finally (* must free this memory *)
                FreeMem(PropList)
              end     (* must free this memory *)
            end   (* nondegenerate *)
      end   (* SetDefaultValues *);

    function TrimPathstr(const S: string; const N: integer): string;
      (* Return a version of S that can be displayed in no more than
      N characters.  The current strategy works mainly from the left,
      changing
           C:\12345\67890\abcdef\ghijklmn\opqrs
      into
           C:\...\67890\abcdef\ghijklmn\opqrs
           C:\...\abcdef\ghijklmn\opqrs
           C:\...\ghijklmn\opqrs
       until it fits.
       *)

      const
        Ellipsis = '...\';

      var
        BSPos: integer;
        Head, Tail: string;

      begin (* TrimPathstr *)
        Result := S;
        if length(S) > N then
          begin (* want to trim *)
            BSPos := pos(BACKSLASH, S);
            if BSPos > 0 then
              begin (* nondegenerate *)
                Head := copy(S, 1, BSPos);
                Tail := copy(S, succ(BSPos), maxint);

                BSPos := pos(BACKSLASH, Tail);
                if BSPos > 0 then
                  begin (* can delete stuff *)
                    Head := Head + Ellipsis;
                    repeat (* delete level *)
                      delete(Tail, 1, BSPos);
                      BSPos := pos(BACKSLASH, Tail)
                    until  (* delete level *) (BSPos = 0) or
                                              (length(Head)+length(Tail) <= N)
                  end   (* can delete stuff *);
                Result := Head + Tail
              end   (* nondegenerate *)
          end   (* want to trim *)
      end   (* TrimPathstr *);
      (* eject *)
    function WhiteSpace(const S: string): boolean;
      (* S is just blanks and tabs *)

      var
        I: integer;

      begin (* WhiteSpace *)
        Result := true;
        for I := 1 to length(S) do
          case S[I] of
            BLANK,
              TAB: continue
              else begin (* non-blank *)
                     Result := false;
                     break
                   end   (* non-blank *)
            end (* case *)
      end   (* WhiteSpace *);

{ tDeleteManager ****************************************************}

    procedure tDeleteManager.DeleteRecordAfterCascade(MasterTable: tTable);
      (* if there are dependent tables, get to them first *)

      var
        I: integer;
        MValue: variant;

      begin (* tDeleteManager.DeleteRecordAfterCascade *)
        for I := 1 to Count do
          with tDeleteRecord(Items[I-1]) do
            if MTable = MasterTable then
              begin (* must delete slave records *)
                MValue := MasterTable.FieldByName(MFieldName).Value;
                while STable.locate(SFieldName, MValue, []) do
                  DeleteRecordAfterCascade(STable)
              end   (* must delete slave records *);
        MasterTable.Delete
      end   (* tDeleteManager.DeleteRecordAfterCascade *);

    procedure tDeleteManager.LogDependency(MasterTable, ChildTable: tTable;
                                           const MasterFieldName,
                                                 ChildFieldName: string);
      (* just store arguments in list *)

      begin (* tDeleteManager.LogDependency *)
        Add(tDeleteRecord.Create(MasterTable, ChildTable,
                                 MasterFieldName, ChildFieldName))
      end   (* tDeleteManager.LogDependency *);
      (* eject *)
{ tDeleteRecord *****************************************************}

    constructor tDeleteRecord.Create(MasterTable, SlaveTable: tTable;
                                     const MFN, SFN: string);
      (* Just store arguments in structure *)

      begin (* tDeleteRecord.Create *)
        inherited Create;

        MTable := MasterTable;
        STable := SlaveTable;
        MFieldName := MFN;
        SFieldName := SFN
      end   (* tDeleteRecord.Create *);

{ tResponsibleList **************************************************}
    procedure tResponsibleList.Clear;
      (* list elements are classes too.  Delete in reverse order to minimize
      time spent reorganizing the list *)

      var
        I: integer;

      begin (* tResponsibleList.Clear *)
        for I := Count downto 1 do
          tObject(Items[I-1]).Free;
        inherited Clear
      end   (* tResponsibleList.Clear *);

    destructor tResponsibleList.Destroy;
      (* responsible for list elements *)

      begin (* tResponsibleList.Destroy *)
        Clear;
        inherited Destroy
      end   (* tResponsibleList.Destroy *);

    function tResponsibleList.Dump: string;
      (* for debugging *)

      var
        I: integer;

      begin (* tResponsibleList.Dump *)
        Result := 'Count = ' + IntToStr(Count);
        for I := 1 to Count do
          Result := Result + NL + tResponsibleObject(Items[I-1]).Dump(I)
      end   (* tResponsibleList.Dump *);
      (* eject *)

{ tRRFCloseBtn ******************************************************}

    procedure tRRFCloseBtn.CloseForm(Sender: tObject);
      (* Close unless user aborts *)

      begin (* tRRFCloseBtn.CloseForm *)
        if Assigned(fOnBeforeClose) then
          fOnBeforeClose(Sender);

        with Owner as tForm do
          begin (* with form *)
            AutoScroll := false;
            if fShrinkAway then
              repeat (* shrink *)
                ScaleBy(100-fShrinkRate, 100);
                Left := (Screen.Width  - Width ) div 2;
                Top  := (Screen.Height - Height) div 2;
                Application.ProcessMessages
              until  (* shrink *) Height < 100;

            Close
          end   (* with form *)
      end   (* tRRFCloseBtn.CloseForm *);

    constructor tRRFCloseBtn.Create(AnOwner: tComponent);
      (* just a BitBtn with a preferred glyph *)

      begin (* tRRFCloseBtn.Create *)
        inherited Create(AnOwner);

        Width := 97;
        Height := 49;
        Caption := '&Close';
        Font.Height := -12;
        Font.Name := 'Arial';
        Font.Style := [fsBold];
        ParentFont := False;
        OnClick := CloseForm;
        fShrinkAway := DefaultShrinkAway;
        SetDefaultValues(self);
        if csDesigning in ComponentState then
          Glyph.LoadFromResourceName(HInstance, 'MYCLOSEBUTTON')
      end   (* tRRFCloseBtn.Create *);
      (* eject *)
    procedure tRRFCloseBtn.Loaded;
      (* wait until resources available *)

      begin (* tRRFCloseBtn.Loaded *)
        inherited Loaded;
        Glyph.LoadFromResourceName(HInstance, 'MYCLOSEBUTTON')
      end   (* tRRFCloseBtn.Loaded *);

    procedure tRRFCloseBtn.SetOnBeforeClose(const Value: tNotifyEvent);
      (* might not really want to close *)

      begin (* tRRFCloseBtn.SetOnBeforeClose *)
        fOnBeforeClose := Value
      end   (* tRRFCloseBtn.SetOnBeforeClose *);

    procedure tRRFCloseBtn.SetShrinkRate(const Value: integer);
      (* Each shrinkage is by (100-this)% *)

      begin (* tRRFCloseBtn.SetShrinkRate *)
        if (Value < 1) or (Value > 99) then
          fShrinkAway := false
        else
          fShrinkRate := Value
      end   (* tRRFCloseBtn.SetShrinkRate *);

{ tRRFDTPicker ******************************************************}
    constructor tRRFDTPicker.Create(AnOwner: tComponent);
      (* Initialize *)

      begin (* tRRFDTPicker.Create *)
        inherited Create(AnOwner);
        fRRFDTAnyFormat := nullstring;
        fRRFDTFormat := dtNONE
      end   (* tRRFDTPicker.Create *);

    procedure tRRFDTPicker.SetRRFDTAnyFormat(const Value: string);
      (* Free text format *)

      begin (* tRRFDTPicker.SetRRFDTAnyFormat *)
        if Value <> fRRFDTAnyFormat then
          begin (* nondegenerate *)
            fRRFDTAnyFormat := Value;
            fRRFDTFormat := dtNONE;
            SendMessage(Handle, DTM_SETFORMAT, MysteryMBZ, longint(PChar(Value)))
          end   (* nondegenerate *)
      end   (* tRRFDTPicker.SetRRFDTAnyFormat *);
      (* eject *)
    procedure tRRFDTPicker.SetRRFDTFormat(const Value: dtformat);
      (* Mainly for design-time use *)

      type
        fmtarray = array[dtformat] of string;

      const
        Fmt: fmtarray = (nullstring, '''Year is ''yyyy',
                         'M/d/yy',   'M/d/yyyy',
                         'MM/dd/yy', 'MM/dd/yyyy',
                         'yyyy/M/d', 'yyyy/MM/dd');

      begin (* tRRFDTPicker.SetRRFDTFormat *)
        if Value <> fRRFDTFormat then
          begin (* nondegenerate *)
            fRRFDTFormat := Value;
            fRRFDTAnyFormat := Fmt[Value];
            if Value <> dtNONE then
              SendMessage(Handle,
                          DTM_SETFORMAT,
                          MysteryMBZ,
                          longint(PChar(fRRFDTAnyFormat)))
          end   (* nondegenerate *)
      end   (* tRRFDTPicker.SetRRFDTFormat *);

{ tRRFLock **********************************************************}

    procedure tRRFLock.Abandon;
      (* Unlock completely *)

      begin (* tRRFLock.Abandon *)
        if Assigned(LockTable) then
          with LockTable do
            if Active then
              begin (* give it up *)
                if State = dsEdit then
                  Cancel;
                Close
              end   (* give it up *)
      end   (* tRRFLock.Abandon *);

    constructor tRRFLock.Create(AnOwner: tComponent);
      (* initialize *)

      begin (* tRRFLock.Create *)
        inherited Create(AnOwner);
        LockTable := tTable.Create(self)
      end   (* tRRFLock.Create *);
      (* eject *)
    destructor tRRFLock.Destroy;
      (* release the lock *)

      begin (* tRRFLock.Destroy *)
        Abandon;
        inherited Destroy
      end   (* tRRFLock.Destroy *);

    function tRRFLock.Lock: boolean;
      (* try to edit *)

      begin (* tRRFLock.Lock *)
        Result := Assigned(fDatabase);
        if Result then
          with LockTable do
            begin (* with LockTable *)
              if not Active then
                begin (* create the file *)
                  DatabaseName := fDatabase.DatabaseName;
                  TableType := ttDefault;
                  TableName := 'RRFLOCK.DB';
                  with FieldDefs do
                    begin (* just one field *)
                      Clear;
                      with AddFieldDef do
                        begin (* specify the field *)
                          Name := 'EditLevel';
                          DataType := ftInteger;
                          Required := false
                        end   (* specify the field *)
                    end   (* just one field *);
                  try    (* this stuff might fail *)
                    if not Exists then
                      CreateTable;
                    Open;
                    Insert;
                    Post
                  except (* this stuff might fail *)
                    Result := false
                  end    (* this stuff might fail *)
                end   (* create the file *);

              if Result then
                if State = dsEdit then
                  FieldByName('EditLevel').AsInteger :=
                               FieldByName('EditLevel').AsInteger + 1
                else
                  try    (* start editing *)
                    Edit;
                    FieldByName('EditLevel').AsInteger := 0
                  except (* start editing *)
                    Result := false
                  end    (* start editing *)
            end   (* with LockTable *)
      end   (* tRRFLock.Lock *);
      (* eject *)
    procedure tRRFLock.SetDataBase(const Value: tDatabase);
      (* property access *)

      begin (* tRRFLock.SetDataBase *)
        if Value <> fDatabase then
          fDatabase := Value
      end   (* tRRFLock.SetDataBase *);

    procedure tRRFLock.Unlock;
      (* release at least one level of lock *)

      begin (* tRRFLock.Unlock *)
        with LockTable do
          if not Active then
            ShowMessage('Attempted to unlock, but lock table inactive')
          else if State <> dsEdit then
            ShowMessage('Attempted to unlock, but lock not set')
          else if FieldByName('EditLevel').AsInteger = 0 then
            Cancel
          else
            FieldByName('EditLevel').AsInteger :=
                   FieldByName('EditLevel').AsInteger - 1
      end   (* tRRFLock.Unlock *);

{ tRRFMemo **********************************************************}

    function tRRFMemo.GetSelectedLine: integer;
      (* Report the line # in which SelStart is found *)

      begin (* tRRFMemo.GetSelectedLine *)
        Result := Perform(EM_LINEFROMCHAR, SelStart, NotUsedMBZ)
      end   (* tRRFMemo.GetSelectedLine *);

    function tRRFMemo.PosLineStart: integer;
      (* Report the character position of the beginning of the
        SelStart-containing line *)

      begin (* tRRFMemo.PosLineStart *)
        Result := SelStart - Perform(EM_LINEINDEX, GetSelectedLine, NotUsedMBZ)
      end   (* tRRFMemo.PosLineStart *);

    function tRRFMemo.SelectedLineIsBlank: boolean;
      (* Line is just blanks and tabs *)

      begin (* tRRFMemo.SelectedLineIsBlank *)
        Result := WhiteSpace(Lines[GetSelectedLine])
      end   (* tRRFMemo.SelectedLineIsBlank *);
      (* eject *)
    procedure tRRFMemo.SetSelectedLine(const Value: integer);
      (* Set SelStart to the beginning of the Valueth line *)

      const
        HorizontalScroll = 0;

      begin (* tRRFMemo.SetSelectedLine *)
        SelStart := 0;
        Perform(EM_LINESCROLL, HorizontalScroll, Value);
        SelStart := Perform(EM_LINEINDEX, Value, NotUsedMBZ);
        Perform(EM_SCROLLCARET, NotUsedMBZ, NotUsedMBZ)
      end   (* tRRFMemo.SetSelectedLine *);

{ tRRFWrapper *******************************************************}

    procedure tRRFWrapper.FlushBuffers(DataSet: tDataSet);
      (* AfterPost of every table *)

      begin (* tRRFWrapper.FlushBuffers *)
        dbiSaveChanges((DataSet as tTable).Handle)
      end   (* tRRFWrapper.FlushBuffers *);

{ global initialization *********************************************}

    initialization
      CountingPages := false;
      RRFWrapper := tRRFWrapper.Create

{ global cleanup ****************************************************}

    finalization
      RRFWrapper.Free

end.

